#include "audio.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "alsa.h"
#ifdef PULSE_AUDIO
#include "pulseaudio.h"
#endif
#include "sounds.h"
#include "msg.h"
#include "util.h"
#include "text_over.h"
#include "eq.h"

typedef struct {
  uint64_t start;
  uint64_t sample;
  int audio_output;
} audio_player_internal_t;

audio_player_t *init_audio_player(uint64_t start, int audio_output)
{
  audio_player_internal_t *ret;
  ret = calloc(1, sizeof(*ret)); if (ret == NULL) abort();
  ret->start = start;
  ret->audio_output = audio_output;
  return ret;
}

void *audio_player(audio_player_t *_a)
{
  audio_player_internal_t *a = _a;
  void *output;
  char *sound;
  int size;
  int latency;
  int paused = 1;
  uint64_t next_sample = 0;
  uint64_t cur_time;
  uint64_t buffer_play_local_time;
  uint64_t local_stream_time;
  char *silence;
  int silence_samples;
  int samples_per_20ms;
  char *audio_data;
  int audio_samples;
  int i;
  float volume = 1;
  float volumeDB = 0;
  eq *e = NULL;
  void *(*output_open)(void);
  int (*output_latency)(void *);
  void (*output_play)(void *, char *, int);
  void (*output_drop)(void *);
  void (*output_drain)(void *);
  int end_id;
  int end_reached = 0;
  int delta_time = 0;

  samples_per_20ms = 20 * 48000 / 1000;
  silence_samples = samples_per_20ms * 2;
  silence = malloc(silence_samples * 2 * sizeof(int16_t));
  if (silence == NULL) abort();

  switch (a->audio_output) {
  case AUDIO_ALSA:
    output_open    = alsa_open;
    output_latency = alsa_latency;
    output_play    = alsa_play;
    output_drop    = alsa_drop;
    output_drain   = alsa_drain;
    break;
#ifdef PULSE_AUDIO
  case AUDIO_PULSE_AUDIO:
    output_open    = pulseaudio_open;
    output_latency = pulseaudio_latency;
    output_play    = pulseaudio_play;
    output_drop    = pulseaudio_drop;
    output_drain   = pulseaudio_drain;
    break;
#endif
  }

  output = output_open();

  while (1) {
    if (paused || has_msg(TO_AUDIO)) {
      arg_t t;
      int msg = get_msg(TO_AUDIO, t);
      switch (msg) {
      case M_PAUSED: paused = 1; break;
      case M_UNPAUSED: paused = 0; a->start = t[0].u; break;
      /* for M_SEEK, video will send M_UNPAUSE when it has an image */
      case M_SEEK: next_sample = 0; output_drop(output); break;
      case M_GAIN:
        volumeDB += t[0].i;
        if (volumeDB < -60) volumeDB = -60;
        if (volumeDB > 60)  volumeDB = 60;
        volume = pow(10, volumeDB/20);
        oclear();
        oprintf("volume %g\n", volumeDB);
        odelay(500);
        send_msg(TO_X, M_REDRAW, 0);
        break;
      case M_SET_EQ: e = t[0].p; break;
      case M_WAITING_END: end_id = t[0].i; end_reached = 1; break;
      case M_SET_LATENCY: delta_time = t[0].i; break;
      }
    }

    if (paused) continue;

    sound = get_sound(&size);
    if (sound == NULL) {
      printf("no sound!\n");
      if (end_reached) {
        arg_t a;
        a[0].i = end_id;
        output_drain(output);
        send_msg(TO_APP, M_END_REACHED, a);
        end_reached = 0;
      }
      usleep(10000);
      continue;
    }

    if (size / 4 != samples_per_20ms) { printf("bad, we deal with 20ms buffer\n"); exit(1); }

try_again:
    latency = output_latency(output);

    cur_time = get_time();
    buffer_play_local_time = cur_time + (uint64_t)latency * 1000 / 48000;
    local_stream_time = next_sample * 1000 / 48000 + a->start + delta_time;

    if (buffer_play_local_time > local_stream_time + 500) {
      /* waaay to late, drop */
      goto dont_play;
    } else
    if (buffer_play_local_time + 500 < local_stream_time) {
      /* waaay to early, play some silence */
      memset(silence, 0, samples_per_20ms * 2 * sizeof(int16_t));
      audio_data = silence;
      audio_samples = samples_per_20ms;
      output_play(output, audio_data, audio_samples);
      goto try_again;
    } else
    if (buffer_play_local_time > local_stream_time + 50) {
      /* too late, smash at most 5% of the samples */
      int16_t *src = (int16_t *)sound;
      int16_t *dst = (int16_t *)silence;
      uint64_t skip = buffer_play_local_time-local_stream_time;
      uint64_t inc;
      int samples = 0;
      if (skip > samples_per_20ms/20) skip = samples_per_20ms/20;
      inc = (samples_per_20ms+skip-1)/skip;
      for (i = 0; i < samples_per_20ms; i++) {
        int j;
        if (i % inc != 0) {
          for (j = 0; j < 2; j++)
            dst[j] = src[j];
          dst += 2;
          samples++;
        }
        src += 2;
      }
      audio_data = silence;
      audio_samples = samples;
    } else
    if (buffer_play_local_time + 50 < local_stream_time) {
      /* too early, duplicate at most 5% of the samples */
      int16_t *src = (int16_t *)sound;
      int16_t *dst = (int16_t *)silence;
      uint64_t more = local_stream_time-buffer_play_local_time;
      uint64_t inc;
      int samples = 0;
      if (more > samples_per_20ms/20) more = samples_per_20ms/20;
      inc = (samples_per_20ms+more-1)/more;
      for (i = 0; i < samples_per_20ms; i++) {
        int j;
        for (j = 0; j < 2; j++) dst[j] = src[j];
        dst += 2;
        samples++;
        if (i % inc == 0) {
          for (j = 0; j < 2; j++) dst[j] = src[j];
          dst += 2;
          samples++;
        }
        src += 2;
      }
      audio_data = silence;
      audio_samples = samples;
    } else {
      /* normal speed */
      audio_data = sound;
      audio_samples = samples_per_20ms;
    }

    /* TODO: optimize */
    for (i = 0; i < audio_samples * 2; i++) {
      int16_t *v = (int16_t *)audio_data;
      float n = v[i] * volume;
      if (n > 32767) n = 32767;
      if (n < -32768) n = -32768;
      v[i] = n;
    }
    if (e != NULL) run_eq(e, (int16_t *)audio_data, audio_samples, 2);
    output_play(output, audio_data, audio_samples);

dont_play:
    free(sound);

    next_sample += size / 4;
  }

  return 0;
}
