#include "eqgui.h"
#include "eq.h"

#include <string.h>
#include <stdlib.h>
#include <cairo/cairo.h>
#include <stdint.h>
#include <math.h>
#include <stdio.h>
#include <complex.h>

#include "conf.h"
#include "x.h"
#include "util.h"

static unsigned char screen[640*352*4];

static float val_to_gui_log(float f, float fmin, float fmax)
{ float a = fmin; float b = log(fmax/fmin); return log(f/a)/b; }
static float gui_to_val_log(float x, float fmin, float fmax)
{ float a = fmin; float b = log(fmax/fmin); return a * exp(b * x); }
static float val_to_gui_gain(float g, float _, float __) { return (g+20)/40; }
static float gui_to_val_gain(float x, float _, float __) { return x*40-20; }

#define KNOB_BUTTON 0
#define ON_OFF_BUTTON 1

typedef struct {
  int x, y;
  int active; /* =1 if mouse over or button is clicked */
  char *label_left, *label_up, *label_right;
  float (*val_to_gui)(float, float, float);
  float (*gui_to_val)(float, float, float);
  float gui;  /* 0: button left, 1: button right, .5: button middle */
  float value_min, value_max;
  int type;
  char *conf_name;
} button;

static button buttons[] = {
  /* frequency buttons */
  { 150, 320, 0, "20", "200Hz", "2000", val_to_gui_log, gui_to_val_log,
    0.5, 20, 2000, KNOB_BUTTON, "eq.0.frequency" },
  { 290, 320, 0, "40", "400Hz", "4000", val_to_gui_log, gui_to_val_log,
    0.5, 40, 4000, KNOB_BUTTON, "eq.1.frequency" },
  { 430, 320, 0, "100", "1000Hz", "10000", val_to_gui_log, gui_to_val_log,
    0.5, 100, 10000, KNOB_BUTTON, "eq.2.frequency" },
  { 570, 320, 0, "200", "2000Hz", "20000", val_to_gui_log, gui_to_val_log,
    0.5, 200, 20000, KNOB_BUTTON, "eq.3.frequency" },
  /* bandwidth buttons */
  { 100, 250, 0, "1/8", "1 octave", "8", val_to_gui_log, gui_to_val_log,
    0.5, .125, 8, KNOB_BUTTON, "eq.0.bandwidth" },
  { 240, 250, 0, "1/8", "1 octave", "8", val_to_gui_log, gui_to_val_log,
    0.5, .125, 8, KNOB_BUTTON, "eq.1.bandwidth" },
  { 380, 250, 0, "1/8", "1 octave", "8", val_to_gui_log, gui_to_val_log,
    0.5, .125, 8, KNOB_BUTTON, "eq.2.bandwidth" },
  { 520, 250, 0, "1/8", "1 octave", "8", val_to_gui_log, gui_to_val_log,
    0.5, .125, 8, KNOB_BUTTON, "eq.3.bandwidth" },
  /* gain buttons */
  { 150, 250, 0, "-20", "0 dB", "+20", val_to_gui_gain, gui_to_val_gain,
    0.5, -20, 20, KNOB_BUTTON, "eq.0.gain" },
  { 290, 250, 0, "-20", "0 dB", "+20", val_to_gui_gain, gui_to_val_gain,
    0.5, -20, 20, KNOB_BUTTON, "eq.1.gain" },
  { 430, 250, 0, "-20", "0 dB", "+20", val_to_gui_gain, gui_to_val_gain,
    0.5, -20, 20, KNOB_BUTTON, "eq.2.gain" },
  { 570, 250, 0, "-20", "0 dB", "+20", val_to_gui_gain, gui_to_val_gain,
    0.5, -20, 20, KNOB_BUTTON, "eq.3.gain" },
  /* on/off buttons */
  { 100, 320, 0, "OFF", NULL, "ON", NULL, NULL, 0, 0, 0, ON_OFF_BUTTON,
    "eq.0.on" },
  { 240, 320, 0, "OFF", NULL, "ON", NULL, NULL, 0, 0, 0, ON_OFF_BUTTON,
    "eq.1.on" },
  { 380, 320, 0, "OFF", NULL, "ON", NULL, NULL, 0, 0, 0, ON_OFF_BUTTON,
    "eq.2.on" },
  { 520, 320, 0, "OFF", NULL, "ON", NULL, NULL, 0, 0, 0, ON_OFF_BUTTON,
    "eq.3.on" },
  /* master gain button */
  { 30, 280, 0, "-20", "0 dB", "+20", val_to_gui_gain, gui_to_val_gain,
    0.5, -20, 20, KNOB_BUTTON, "eq.gain" },
  { 0 }
};

static void update_buttons(void)
{
  extern eq e;
  int i;
  for (i = 0; buttons[i].x; i++) buttons[i].active = 0;
  /* freq */
  for (i = 0; i < 4; i++)
    buttons[i].gui = val_to_gui_log(e.frequency[i],
        buttons[i].value_min, buttons[i].value_max);
  /* bandwidth */
  for (i = 0; i < 4; i++)
    buttons[4+i].gui = val_to_gui_log(e.bandwidth[i],
        buttons[4+i].value_min, buttons[4+i].value_max);
  /* gain */
  for (i = 0; i < 4; i++)
    buttons[8+i].gui = val_to_gui_gain(e.gain[i], 0, 0);
  /* on/off */
  for (i = 0; i < 4; i++)
    buttons[12+i].gui = e.section[i];
  /* master gain */
  buttons[16].gui = val_to_gui_gain(e.filter_gain, 0, 0);
}

static void convert(void)
{
  uint32_t *p = (uint32_t *)screen;
  unsigned char *to = screen;
  uint32_t x;
  int i;
  for (i = 0; i < 640*352; i++, p++) {
    x = *p;
    *to++ = (x >> 16) & 255;
    *to++ = (x >> 8) & 255;
    *to++ = (x) & 255;
    *to++ = (x >> 24) & 255;
  }
}

static double _a[4], _s1[4], _s2[4];

static void setup(int i, float f0, float dBgain, float bw, float fs)
{
  double w0 = 2 * M_PI * f0 / fs;
  double g = pow(10, dBgain/20);
  double b = bw * 7 * f0/fs / sqrt(g);
  _a[i] = (g-1) / 2;
  _s1[i] = -cos(w0);
  _s2[i] = (1-b)/(1+b);
}

static void compute_init(void)
{
  extern eq e;
  int i;
  for (i = 0; i < 4; i++)
    setup(i, e.frequency[i], e.gain[i], e.bandwidth[i], e.samplerate);
}

static float compute_one_eq(int i, float f, float fs)
{
  complex z = cexp(-I * f / fs * 2 * M_PI);
  complex d;
  double K = _a[i] * 2 + 1;
  double b = _s1[i];
  double a = _s2[i];
  complex Az = (z * z + b * (1 + a) * z + a) /
               (1 + b * (1 + a) * z + a * z * z);
  d = .5 * (1 + Az) + .5 * K * (1 - Az);
  return 20*log10(cabs(d));
}

static float compute_eq(int i)
{
  extern eq e;
  float f;
  float ret = 0;
  f = 20 * exp(log(20000./20.)/640. * i);
  for (i = 0; i < 4; i++)
    if (buttons[i+12].gui == 1) ret += compute_one_eq(i, f, e.samplerate);
  ret += e.filter_gain;
  /* 20   0
   * -20  199
   * x    y
   * (x-20)/-40 = y/199
   */
  return (20-ret)/40*199;
}

static void draw_button(cairo_t *c, button *b)
{
  cairo_text_extents_t ext;

  if (b->type == ON_OFF_BUTTON) {
    int x = b->x-10;
    if (b->active) {
      cairo_set_source_rgb(c, .6, .4, .4);
      cairo_move_to(c, x, b->y);
      cairo_arc(c, x, b->y, 10, 0, 2*M_PI);
      cairo_fill(c);
    }
    if (b->gui == 0)
      cairo_set_source_rgb(c, 1, .2, .2);
    else
      cairo_set_source_rgb(c, .2, 1, .2);
    cairo_rectangle(c, x-5, b->y-5, 10, 10);
    cairo_fill(c);
    cairo_set_source_rgba(c, .4, 1, 1, .5);
    cairo_move_to(c, x+8, b->y+4);
    cairo_show_text(c, b->gui==0 ? b->label_left : b->label_right);
    return;
  }

  /* KNOB_BUTTON */
  if (b->active) {
    cairo_set_source_rgb(c, .6, .4, .4);
    cairo_move_to(c, b->x, b->y);
    cairo_arc(c, b->x, b->y, 20, 0, 2*M_PI);
    cairo_fill(c);
  }
  cairo_set_source_rgba(c, .4, 1, 1, .5);
  cairo_move_to(c, b->x, b->y);
  cairo_arc(c, b->x, b->y, 10, 3*M_PI/4, M_PI/4);
  cairo_close_path(c);
  cairo_fill(c);
  cairo_new_sub_path(c);
  cairo_arc(c, b->x, b->y, 12, 3*M_PI/4, 3*M_PI/4 + b->gui * 2 * 3*M_PI/4);
  cairo_line_to(c, b->x, b->y);
  cairo_set_source_rgb(c, 1, .6, .6);
  cairo_fill_preserve(c);
  cairo_set_source_rgb(c, 1, .2, .2);
  cairo_stroke(c);

  cairo_set_source_rgba(c, .4, 1, 1, .5);
  if (b->label_left != NULL) {
    cairo_text_extents(c, b->label_left, &ext);
    cairo_move_to(c, b->x-10-ext.width, b->y+20);
    cairo_show_text(c, b->label_left);
  }
  if (b->label_right != NULL) {
    cairo_move_to(c, b->x+5, b->y+20);
    cairo_show_text(c, b->label_right);
  }
  if (b->label_up != NULL) {
    cairo_text_extents(c, b->label_up, &ext);
    cairo_move_to(c, b->x-ext.width/2, b->y-17);
    cairo_show_text(c, b->label_up);
  }
}

static void paint(eqgui *e)
{
  cairo_t *c;
  float v;
  int i;
  float freq;

  c=cairo_create(e->cairo_surface);

  /* clear screen */
  memset(screen, 0, 640*352*4);
  cairo_rectangle(c, 0, 0, 640, 352);
  cairo_set_source_rgba(c, 0, 0, 0, 0.8);
  cairo_fill(c);

  cairo_rectangle(c, 0, 0, 640, 200);
  cairo_clip(c);

  /* plot curve */
  compute_init();
  v = compute_eq(0);
  cairo_move_to(c, -10, v);
  for (i = 1; i < 640; i++) {
    v = compute_eq(i);
    cairo_line_to(c, i, v);
  }
  cairo_line_to(c, 640+10, v);
  cairo_line_to(c, 640+10, 100);
  cairo_line_to(c, 640+10, 400);
  cairo_line_to(c, -10, 400);
  cairo_set_source_rgb(c, .4, .4, 1);
  cairo_fill_preserve(c);
  cairo_set_source_rgb(c, .2, .2, 1);
  cairo_stroke(c);

  cairo_reset_clip(c);

  /* draw -20 / 0 / 20 dB horizontal lines */
  cairo_set_source_rgba(c, .4, 1, 1, .5);
  cairo_move_to(c, 0, 1);
  cairo_line_to(c, 640, 1);
  cairo_move_to(c, 0, 100);
  cairo_line_to(c, 640, 100);
  cairo_move_to(c, 0, 200);
  cairo_line_to(c, 640, 200);

  cairo_set_line_width(c, 2);
  /* draw octave vertical lines */
  for (freq = 20; freq <= 20000; freq *= 2) {
    int delta = freq==20;
    i = log(freq / 20.) / (log(20000./20.)/639.);
    cairo_move_to(c, i+delta, 0);
    cairo_line_to(c, i+delta, 200);
  }
  cairo_move_to(c, 639, 0);
  cairo_line_to(c, 639, 200);
  cairo_stroke(c);

  cairo_set_font_size(c, 10);
  cairo_move_to(c, 1, 212);   cairo_show_text(c, "20Hz");
  cairo_move_to(c, 609, 212); cairo_show_text(c, "20000");
  for (freq = 40; freq <= 20000; freq *= 2) {
    cairo_text_extents_t ext;
    char s[64];
    sprintf(s, "%g", freq);
    cairo_text_extents(c, s, &ext);
    i = log(freq / 20.) / (log(20000./20.)/639.);
    cairo_move_to(c, i-ext.width/2, 212);
    cairo_show_text(c, s);
  }

  cairo_move_to(c, 3, 12);   cairo_show_text(c, "20");
  cairo_move_to(c, 3, 98);   cairo_show_text(c, "0dB");
  cairo_move_to(c, 3, 198);   cairo_show_text(c, "-20");

  /* draw buttons */
  for (i = 0; buttons[i].x; i++) draw_button(c, &buttons[i]);

  /* print value of current active button, if any */
  if (e->active_button != -1 && buttons[e->active_button].type==KNOB_BUTTON) {
    button *b = &buttons[e->active_button];
    char s[128];
    cairo_text_extents_t ext;
    sprintf(s, "%g", b->gui_to_val(b->gui, b->value_min, b->value_max));
    cairo_text_extents(c, s, &ext);
    cairo_set_source_rgb(c, 1, .6, .6);
    cairo_move_to(c, b->x-ext.width/2, b->y-30);
    cairo_show_text(c, s);
  }

  cairo_destroy(c);

  convert();
}

void eqgui_init(eqgui *e)
{
  memset(e, 0, sizeof(*e));
  e->cairo_surface = cairo_image_surface_create_for_data(screen,
      CAIRO_FORMAT_ARGB32, 640, 352, 640*4);
  e->active_button = -1;
  e->clicked_button = -1;
}

void eqgui_enter(eqgui *e)
{
  update_buttons();
  e->visible = 1;
  e->active_button = -1;
  e->clicked_button = -1;
  //paint(e);
  e->dirty = 1;
}

static void transform_coordinates(int x, int y,
    int window_width, int window_height, double *xf, double *yf,
    double *xminr, double *yminr)
{
  int width, height;
  double scale, sw, sh;
  double xmin, xmax, ymin, ymax;

  /* might be faster computation somewhere */
  width = 640;
  height = 352;
  scale = window_width / (double)width;
  if (scale * height > window_height)
    scale = window_height / (double)height;
  sw = scale*width;
  sh = scale*height;

  xmin = (window_width-sw)/2;
  xmax = xmin + sw;
  ymin = (window_height-sh)/2;
  ymax = ymin + sh;

  /* xmin -> 0
   * xmax -> 640
   * x    -> xi
   * (x-xmin)/(xmax-xmin) = xi/640
   * xi = (x-xmin)/(xmax-xmin) * 640
   */
  *xf = (x-xmin)/(xmax-xmin) * 640;
  *yf = (y-ymin)/(ymax-ymin) * 352;

  if (xminr != NULL) *xminr = xmin;
  if (yminr != NULL) *yminr = ymin;
}

static int get_button(eqgui *e, int x, int y,
                      int window_width, int window_height)
{
  int i;
  double xf, yf, xmin, ymin;

  if (e->clicked_button != -1) return e->clicked_button;

  transform_coordinates(x, y, window_width, window_height,
                        &xf, &yf, &xmin, &ymin);

  for (i = 0; buttons[i].x; i++) {
    if (buttons[i].type == ON_OFF_BUTTON) {
      xmin = buttons[i].x-10;
      ymin = buttons[i].y;
      if ((xmin-xf)*(xmin-xf)+(ymin-yf)*(ymin-yf) < 10*10) return i;
      continue;
    }
    /* KNOB_BUTTON */
      xmin = buttons[i].x;
      ymin = buttons[i].y;
      if ((xmin-xf)*(xmin-xf)+(ymin-yf)*(ymin-yf) < 20*20) return i;
  }

  return -1;
}

int eqgui_button_press(eqgui *e, int x, int y, int but,
                       int width, int height)
{
  int button;

  if (but != 1) return OVER_NOTHING;

  button = get_button(e, x, y, width, height);
  if (e->active_button != -1) buttons[e->active_button].active = 0;
  if (button != -1) {
    buttons[button].active = 1;
    e->clicked_button = button;
  }

  //paint(e);
  e->dirty = 1;

  return OVER_REDRAW;
}

int eqgui_button_release(eqgui *eg, int x, int y, int but,
                         int width, int height)
{
  extern eq e;
  int button;

  if (but != 1) return OVER_NOTHING;

  eg->clicked_button = -1;
  button = get_button(eg, x, y, width, height);

  if (eg->active_button != -1) {
    if (buttons[eg->active_button].type == ON_OFF_BUTTON) {
      if (button == eg->active_button) {
        e.section[button % 4] = 1 - e.section[button % 4];
        buttons[button].gui = 1 - buttons[button].gui;
        store_conf_float(buttons[button].conf_name, e.section[button % 4]);
      }
    }
  }

  if (eg->active_button != -1) buttons[eg->active_button].active = 0;
  if (button != -1) buttons[button].active = 1;
  eg->active_button = button;

  //paint(eg);
  eg->dirty = 1;

  return OVER_REDRAW;
}

static float *get_eq_var(int button)
{
  extern eq e;
  switch (button) {
  case  0: case  1: case  2: case  3: return &e.frequency[button % 4];
  case  4: case  5: case  6: case  7: return &e.bandwidth[button % 4];
  case  8: case  9: case 10: case 11: return &e.gain[button % 4];
  case 16: return &e.filter_gain;
  }
  return NULL;
}

int eqgui_pointer_move(eqgui *eg, int x, int y, int width, int height)
{
  int button;
  int ret = OVER_NOTHING;

  button = get_button(eg, x, y, width, height);
  if (button != eg->active_button) ret = OVER_REDRAW;
  if (eg->active_button != -1) buttons[eg->active_button].active = 0;
  if (button != -1) buttons[button].active = 1;
  eg->active_button = button;

  if (eg->active_button != -1 && eg->active_button == eg->clicked_button) {
    if (buttons[button].type == KNOB_BUTTON) {
      double angle;
      double val, eqval;
      double xf, yf;
      double x_center, y_center;
      transform_coordinates(x, y, width, height, &xf, &yf, NULL, NULL);
      x_center = buttons[button].x;
      y_center = buttons[button].y;
      angle = -atan2(x_center-xf, y_center-yf);
      val = (angle - (-M_PI/2 - M_PI/4)) / (1.5 * M_PI);
      if (val < 0) val = 0; else if (val > 1) val = 1;
      eqval = buttons[button].gui_to_val(val,
                  buttons[button].value_min, buttons[button].value_max);
      *get_eq_var(button) = eqval;
      if (buttons[button].gui != val) {
        store_conf_float(buttons[button].conf_name, eqval);
        ret = OVER_REDRAW;
      }
      buttons[button].gui = val;
    }
  }

  if (ret == OVER_REDRAW) eg->dirty = 1; //paint(eg);

  return ret;
}

int eqgui_key_press(eqgui *e, int k)
{
  if (k == KEY_ESC) { e->visible = 0; return OVER_QUIT; }
  return OVER_NOTHING;
}

char *eqgui_get_screen(eqgui *e)
{
  if (e->dirty) { paint(e); e->dirty = 0; }
  if (e->visible) return (char *)screen;
  return NULL;
}
