#include "x.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/glx.h>
#include <unistd.h>
#include <poll.h>

#include "thread.h"
#include "images.h"
#include "msg.h"
#include "text_over.h"
#include "eqgui.h"
#include "util.h"

typedef struct {
  Display *d;
  Window w;
  GLXContext glc;
  int fullscreen;
  XVisualInfo *vi;
  GLuint texture[2];
  Colormap cmap;
  XSetWindowAttributes swa;
  int image_width;
  int image_height;
  int display_width;
  int display_height;
  int fps;
  int pipe[2];  /* 0: read   1: write */
  eqgui *eg;
} x_internal_t;

static void select_input(Display *d, Window w)
{
  XSelectInput(d, w, KeyPressMask | EnterWindowMask | PointerMotionMask |
               StructureNotifyMask | ExposureMask | ButtonPressMask |
               ButtonReleaseMask);
}


/* Where resizing from full screen to window, there was sometimes
 * a blocked 'poll' when I called glClear which calls xxx
 * calling _XReply calling xxx calling xcb_wait_for_reply64 calling xxx
 * calling poll and this poll waits forever for whatever reason.
 * Waiting for exposure here seems to solve this problem.
 * Not sure it's a bug on my side (maybe weird interactions with
 * the window manager? fvwm in my case)...
 */
void map_window_and_wait(Display *d, Window w)
{
  XEvent e;
  XMapWindow(d, w);
  XWindowEvent(d, w, ExposureMask, &e);
  if (e.type != Expose) abort();
  XPutBackEvent(d, &e);
}

void hide_pointer(Display *d, Window w)
{
  Cursor invisibleCursor;
  Pixmap bitmapNoData;
  XColor black;
  static char noData[] = { 0,0,0,0,0,0,0,0 };
  black.red = black.green = black.blue = 0;

  bitmapNoData = XCreateBitmapFromData(d, w, noData, 8, 8);
  invisibleCursor = XCreatePixmapCursor(d, bitmapNoData, bitmapNoData,
                                       &black, &black, 0, 0);
  XDefineCursor(d, w, invisibleCursor);
  XFreeCursor(d, invisibleCursor);
  XFreePixmap(d, bitmapNoData);
}

void show_pointer(Display *d, Window w)
{
  XUndefineCursor(d, w);
}

/* the creation of the GL context has to be done by the X thread (if I
 * understand properly) so we have both init_display() called by the app
 * and internal_init_display() called by the X thread
 */
static void internal_init_display(x_internal_t *x)
{
  int display_width;
  int display_height;
  Display *d;
  Window w;
  GLXContext glc;
  XVisualInfo *vi;
  GLuint texture[2];
  Colormap cmap;
  XSetWindowAttributes swa;
  GLint att[] = { GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None };

  d = XOpenDisplay(0); if (d==NULL) abort();

  vi = glXChooseVisual(d, DefaultScreen(d), att);
  if (vi == NULL) { fprintf(stderr, "gl error\n"); abort(); }

  cmap = XCreateColormap(d, RootWindow(d, vi->screen), vi->visual, AllocNone);
  swa.colormap = cmap;

  if (!x->fullscreen) {
    w = XCreateWindow(d, RootWindow(d, vi->screen),
            0, 0, x->image_width, x->image_height, 0,
            vi->depth, InputOutput, vi->visual,
            CWColormap, &swa);
    display_width = x->image_width;
    display_height = x->image_height;
  } else {
    Atom wm_state;
    Atom wm_fullscreen;
    XSizeHints hints;
    XWindowAttributes wa;
    XGetWindowAttributes(d, DefaultRootWindow(d), &wa);
    w = XCreateWindow(d, DefaultRootWindow(d), 0, 0, wa.width, wa.height, 0,
                      vi->depth, InputOutput, vi->visual,
                      CWColormap, &swa);
    display_width = wa.width;
    display_height = wa.height;
    wm_state   = XInternAtom (d, "_NET_WM_STATE", True );
    wm_fullscreen = XInternAtom (d, "_NET_WM_STATE_FULLSCREEN", True );
    XChangeProperty(d, w, wm_state, XA_ATOM, 32,
                    PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
//    XMoveWindow(d, w, 0, 0);
    hints.flags = USPosition;
    hints.x = hints.y = 0;
    XSetWMNormalHints(d, w, &hints);
  }

  select_input(d, w);
  map_window_and_wait(d, w);

  glc = glXCreateContext(d, vi, NULL, GL_TRUE);
  if (glXMakeCurrent(d, w, glc) == False) abort();

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glGenTextures(2, texture);
  glEnable(GL_TEXTURE_2D);
  //glActiveTexture(GL_TEXTURE0+0);
  glBindTexture(GL_TEXTURE_2D, texture[0]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  /* not sure about this clamping */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);
  glBindTexture(GL_TEXTURE_2D, texture[1]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  /* not sure about this clamping */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);

  //XFlush(d);

  x->d = d;
  x->w = w;
  x->glc = glc;
  x->vi = vi;
  x->texture[0] = texture[0];
  x->texture[1] = texture[1];
  x->cmap = cmap;
  x->swa = swa;
  x->display_width = display_width;
  x->display_height = display_height;
}

x_t *init_display(int width, int height, int fps, int fullscreen, eqgui *eg)
{
  x_internal_t *ret;

  ret = calloc(1, sizeof(*ret)); if (ret == NULL) abort();
  ret->fullscreen = fullscreen;
  ret->image_width = width;
  ret->image_height = height;
  ret->fps = fps;
  ret->eg = eg;

  if (pipe(ret->pipe)) abort();

  return ret;
}

void display_image(x_internal_t *x, char *image, eqgui *eg)
{
  double scale;
  double sw, sh;
  int width, height;

  glXMakeContextCurrent(x->d, x->w, x->w, x->glc);
  glViewport(0, 0, x->display_width, x->display_height);
  glClear(GL_COLOR_BUFFER_BIT);

  if (image == NULL) goto no_background;

  scale = x->display_width / (double)x->image_width;
  if (scale * x->image_height > x->display_height)
    scale = x->display_height / (double)x->image_height;
  sw = scale * x->image_width;
  sh = scale * x->image_height;
  glViewport((x->display_width-sw)/2, (x->display_height-sh)/2, sw, sh);
  //glActiveTexture(GL_TEXTURE0+0);
  glBindTexture(GL_TEXTURE_2D, x->texture[0]);

  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, x->image_width, x->image_height, 0,
               GL_RGBA, GL_UNSIGNED_BYTE, image);

  glLoadIdentity();
  glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex3f( 1, -1, 0);
  glTexCoord2f(0, 1); glVertex3f(-1, -1, 0);
  glTexCoord2f(0, 0); glVertex3f(-1,  1, 0);
  glTexCoord2f(1, 0); glVertex3f( 1,  1, 0);
  glEnd();

no_background:
  if (eg->visible)
    image = eqgui_get_screen(eg);
  else
    image = get_text_over();
  if (image == NULL) goto no_text_over;

  /* size of text over is 640x352 */
  width = 640;
  height = 352;

  scale = x->display_width / (double)width;
  if (scale * height > x->display_height)
    scale = x->display_height / (double)height;
  sw = scale*width;
  sh = scale*height;
  glViewport((x->display_width-sw)/2, (x->display_height-sh)/2, sw, sh);
  //glActiveTexture(GL_TEXTURE0+1);
  glBindTexture(GL_TEXTURE_2D, x->texture[1]);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
               GL_RGBA, GL_UNSIGNED_BYTE, image);
  glLoadIdentity();
  glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex3f( 1, -1, 0);
  glTexCoord2f(0, 1); glVertex3f(-1, -1, 0);
  glTexCoord2f(0, 0); glVertex3f(-1,  1, 0);
  glTexCoord2f(1, 0); glVertex3f( 1,  1, 0);
  glEnd();

no_text_over:
  glXSwapBuffers(x->d, x->w);

int gl_error = glGetError( );
if (gl_error != GL_NO_ERROR) printf("err %d\n", gl_error);
}

/* this thread receives msg sent to TO_X and forwards them to TO_X_INTERNAL
 * writing to the pipe to signal X thread to wake up for a messgae
 * (dirty? dirty yourself!)
 */
void *xthread_msg(x_t *_x)
{
  x_internal_t *x = _x;

  while (1) {
    arg_t a;
    int msg = get_msg(TO_X, a);
    send_msg(TO_X_INTERNAL, msg, a);
    char v = 1;
    if (write(x->pipe[1], &v, 1) != 1) abort();
  }
}

/* code to toggle fullscreen more or less stolen from xterm */
#define _NET_WM_STATE_REMOVE 0
#define _NET_WM_STATE_ADD    1

static void toggle_fullscreen(x_internal_t *x)
{
  /* that works for me (fvwm), it might well fail for you */
  XEvent e;
  Atom atom_fullscreen = XInternAtom(x->d, "_NET_WM_STATE_FULLSCREEN", False);
  Atom atom_state = XInternAtom(x->d, "_NET_WM_STATE", False);
  memset(&e, 0, sizeof(e));
  e.xclient.type = ClientMessage;
  e.xclient.message_type = atom_state;
  e.xclient.display = x->d;
  e.xclient.window = x->w;
  e.xclient.format = 32;
  e.xclient.data.l[0] = x->fullscreen ? _NET_WM_STATE_REMOVE : _NET_WM_STATE_ADD;
  e.xclient.data.l[1] = atom_fullscreen;
  XSendEvent(x->d, DefaultRootWindow(x->d), False, SubstructureRedirectMask, &e);
  x->fullscreen = 1 - x->fullscreen;
}

static void set_window_size_from_video(x_internal_t *x)
{
  if (glXMakeCurrent(x->d, None, NULL) == False) abort();
  XDestroyWindow(x->d, x->w);

  x->w = XCreateWindow(x->d, DefaultRootWindow(x->d), 0, 0, x->image_width, x->image_height, 0,
                       x->vi->depth, InputOutput, x->vi->visual,
                       CWColormap, &x->swa);
  x->fullscreen = 0;
  select_input(x->d, x->w);
  map_window_and_wait(x->d, x->w);
  if (glXMakeCurrent(x->d, x->w, x->glc) == False) abort();
  //XFlush(d);
}

/* this thread is responsible to keep the screen 'on' (no screen saver,
 * no blanking). It works on my setup, maybe not on every setup.
 */
void *screen_always_on(void *_)
{
  int paused = 0;
  arg_t a;
  /* force screen 'on' every minute when not paused */
  int delay_ms = 60 * 1000;

  new_timer(delay_ms, TO_SCREEN_ON, M_TIMEOUT, a);

  while (1) {
    int msg = get_msg(TO_SCREEN_ON, a);
    switch (msg) {
    case M_PAUSED: paused = 1; break;
    case M_UNPAUSED: paused = 0; break;
    case M_TIMEOUT:
      if (!paused) {
        /* calling those two commands forces the screen 'on' on my system */
        system("xset dpms force on");
        system("xset s reset");
      }
      new_timer(delay_ms, TO_SCREEN_ON, M_TIMEOUT, a);
      break;
    }
  }
  return 0;
}

void *xthread(x_t *_x)
{
  char *image = NULL;
  x_internal_t *x = _x;
  arg_t a;
  int immediate_display;
  struct pollfd fds[2];

  new_thread(screen_always_on, NULL, "screen on");

  internal_init_display(x);
  send_msg(TO_X_WAIT, M_UNPAUSED, 0);

  while (1) {
    immediate_display = 0;

    while (has_msg(TO_X_INTERNAL)) {
      int msg_type = get_msg(TO_X_INTERNAL, a);
      switch (msg_type) {
      case M_SHOW_IMAGE: {
        char *new_image = a[0].p;
        if (new_image != NULL) {
          display_image(x, new_image, x->eg);
          free(image);
          image = new_image;
        }
        /* in some cases, when you resize the window for example, video images
         * will accumulate in the TO_X_INTERNAL message queue, which may
         * explode memory usage, so let's the video thread wait for us to
         * consume the image before sending the next
         * (the video buffer will block in the meantime, no memory madness)
         */
        send_msg(TO_VIDEO_FROM_X, M_CONSUMED, 0);
        break;
      }
      case M_REDRAW: immediate_display = 1; break;
      case M_HIDE_POINTER: hide_pointer(x->d, x->w); break;
      case M_SHOW_POINTER: show_pointer(x->d, x->w); break;
      case M_TOGGLE_FULLSCREEN: toggle_fullscreen(x); break;
      case M_SIZE_FROM_VIDEO: set_window_size_from_video(x); break;
      default: abort();
      }
    }

    XFlush(x->d);

    while (XPending(x->d)) {
      XEvent ev;
      XNextEvent(x->d, &ev);
      switch (ev.type) {
      case KeyPress: {
        char buf[65];
        KeySym ks;
        int k = -1;
        XLookupString(&ev.xkey, buf, 64, &ks, NULL);
        switch (ks) {
          case ' ':
          case 'a' ... 'z':
          case 'A' ... 'Z':
          case '0' ... '9':  k = ks; break;
          case XK_Escape:    k = KEY_ESC; break;
          case XK_Left:      k = KEY_LEFT; break;
          case XK_Right:     k = KEY_RIGHT; break;
          case XK_Up:        k = KEY_UP; break;
          case XK_Down:      k = KEY_DOWN; break;
          case XK_Page_Up:   k = KEY_PAGE_UP; break;
          case XK_Page_Down: k = KEY_PAGE_DOWN; break;
        }
        a[0].i = k;
        if (k != -1) send_msg(TO_APP, M_KEYPRESS, a);
        break;
      }
      case MotionNotify:
        a[0].i = ev.xmotion.x;
        a[1].i = ev.xmotion.y;
        a[2].i = x->display_width;
        a[3].i = x->display_height;
        send_msg(TO_APP, M_POINTER_MOVE, a);
        break;
      case EnterNotify:
        a[0].i = ev.xcrossing.x;
        a[1].i = ev.xcrossing.y;
        a[2].i = x->display_width;
        a[3].i = x->display_height;
        send_msg(TO_APP, M_POINTER_MOVE, a);
        break;
      case ButtonPress:
      case ButtonRelease:
        a[0].i = ev.xbutton.x;
        a[1].i = ev.xbutton.y;
        a[2].i = x->display_width;
        a[3].i = x->display_height;
        a[4].i = ev.xbutton.button;
        send_msg(TO_APP, ev.type == ButtonPress ? M_POINTER_CLICK : M_POINTER_UNCLICK, a);
        break;
      case Expose: immediate_display = 1; break;
      case ConfigureNotify:
        x->display_width = ev.xconfigure.width;
        x->display_height = ev.xconfigure.height;
        break;
      }
    }
    if (immediate_display) {
      display_image(x, image, x->eg);
    }

    /* not sure about this - the idea is to not sleep on ConnectionNumber
     * if some events are in the event queue (if I understand correctly the
     * X event stuff)
     */
    XFlush(x->d);
    if (XPending(x->d)) continue;

    fds[0].fd = x->pipe[0];
    fds[0].events = POLLIN;
    fds[1].fd = ConnectionNumber(x->d);
    fds[1].events = POLLIN;

    if (poll(fds, 2, -1) == -1) { perror("poll"); exit(1); }
    if (fds[0].revents & POLLIN) {
      char b[1024];
      read(x->pipe[0], b, 1024);
    }
  }

  return 0;
}

/* called by app to wait for X initialization (window creation, blabla) */
void wait_for_x(void)
{
  arg_t a;
  int msg_type = get_msg(TO_X_WAIT, a);
  if (msg_type != M_UNPAUSED) {
    fprintf(stderr, "fatal: bad msg received in wait_for_x()\n");
    exit(1);
  }
}
